import java.net.*;
import java.io.*;

public class Connection {
	boolean busy;
	
	Socket socket;				// Connection's socket
	
	InputStream inStream;		// InputStreams for receiving data
	DataInputStream inData;
	
	OutputStream outStream;		// Output Streams for sending data
	DataOutputStream outData;
	
	/** Constructor for Connection Object
	 * 
	 * @param port	Port on which the connection acts
	 * 
	 * @throws IOException	Thrown on failure to initiate data streams
	 */
	public Connection() throws IOException{
		this.busy = false;
	}
	
	/**
	 *  Setter for Busy boolean for use as an additional blocker
	 */
	public void setBusy() {
		this.busy = true;
	}
	
	/**
	 *  Getter for additional Busy blocker
	 * @return
	 */
	public boolean isBusy() {
		return this.busy;
	}
	
	/** Constructor for Connection Object (Client Version)
	 * 
	 * @param addr	Server address
	 * @param port  Server port
	 * 
	 * @throws IOException	Thrown on failure to initiate data streams
	 */
	public Connection (String addr, int port) throws IOException{
		this.socket = new Socket(addr, port);
		
		initStreams();
		this.socket.setSoTimeout(5000);  // Timeout set to 5s
	}
	
	/** Setting for sockets used by ServerConnection for opening new sockets
	 * 
	 * @param s				The new socket
	 * @throws IOException	Thrown on socket errors
	 */
	public void newSocket (Socket s) throws IOException{
		this.socket = s;
		
		initStreams();
		this.socket.setSoTimeout(5000);  // Timeout set to 5s
	}
	
	/**
	 * Private method which initialises the stream/data streams based on socket
	 * @throws IOException Thrown on Socket errors
	 */
	private void initStreams() throws IOException {
		this.inStream = this.socket.getInputStream();
		this.inData = new DataInputStream(inStream);
		
		this.outStream = this.socket.getOutputStream();
		this.outData = new DataOutputStream(outStream);
	}
	
	/**	Method for sending a string over the connection 
	 * 
	 * @param output	String to be sent
	 * @throws IOException	Thrown on failure to send data
	 */
	public void sendText(String output) throws IOException {
		this.outData.writeUTF(output);
	}
	
	/** Method for receiving a string over the connection
	 * 
	 * @return	String received over connection
	 * @throws IOException	Thrown on timeout or failure to get data
	 */
	public String recvText() throws IOException {
		return new String(inData.readUTF());
	}
	
	/**	Method for sending an int over the connection 
	 * 
	 * @param output		Integer to be sent
	 * @throws IOException	Thrown on socket error
	 */
	public void sendInt(int output) throws IOException {
		this.outData.writeInt(output);
	}
	
	/** Method for receiving an int over the connection
	 * 
	 * @return	Integer		Integer received
	 * @throws IOException	Thrown on timeout or failure to get data
	 */
	public int recvInt() throws IOException {
		return inData.readInt();
	}

	/**	Method for sending a file over the connection 
	 * 
	 * @param filename		Filename to be read from to send
	 * @throws IOException	Thrown on failure to send data
	 */
	public void sendFile(String filename) throws IOException {
		
		DataInputStream inFileData;
		DataOutputStream outFileData;
		
		inFileData  = new DataInputStream(new FileInputStream(filename));
		outFileData = outData;
		
		int size = inFileData.available();
		outData.writeInt(size);
		
		fileTransfer(inFileData, outFileData, size);
		
		inFileData.close();
	}
	
	/** Method for receiving a file over the connection
	 * 
	 * @param filename		Filename to write to
	 * @throws IOException	Thrown on socket error
	 */
	public void recvFile(String filename) throws IOException {
		DataInputStream inFileData;
		DataOutputStream outFileData;
		
		int size = inData.readInt();
		inFileData = inData;
		outFileData = new DataOutputStream(new FileOutputStream(filename));
		
		fileTransfer(inFileData, outFileData, size);
		
		outFileData.close();
	}
	
	/**	Private method that reads from an input stream and writes it to an output stream
	 * 
	 * @param input			Input stream for reading
	 * @param output		Output stream for writing
	 * @param remaining		Number of bytes remaining to be sent
	 * @throws IOException	Thrown on socket failures
	 */
	private void fileTransfer(DataInputStream input, DataOutputStream output, int remaining) throws IOException {
		int chunksize = 1024, writecount;
		byte bytes[] = new byte[chunksize];
		while (remaining > 0)
		{
			writecount = (remaining >= chunksize ? chunksize : remaining);
			input.read(bytes, 0, writecount);
			output.write(bytes, 0, writecount);
			remaining -= writecount;
		}
		this.busy = false;
	}
	
	/**	Closes connection
	 * 
	 */
	public void closeSocket() {
		try {
			this.inData.close();
			this.inStream.close();
			
			this.outData.close();
			this.outStream.close();
			
			this.socket.close();
		} catch (IOException e) {
			System.err.println(e.getMessage());
		}
	}
}