import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.*;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import javax.swing.*;
import javax.swing.table.*;

public class GUI extends JFrame {
	private static final long serialVersionUID = -9120610634367308325L;
	
	private WorkerManager workers;
	private JobManager jobs;
	private ApplicationManager applications;
	
	private JTable jtWorkers;
	private JTable jtJobs;
	private JTable jtApplications;
	
	/**
	 * GUI Constructor.
	 * @param workers The WorkerManager created by the Master class.
	 * @param jobsThe JobManager created by the Master class.
	 */
	public GUI(WorkerManager workers, JobManager jobs) {
		applications = new ApplicationManager();
		this.workers = workers;
		this.jobs = jobs;
		initUI();
	}
	
	/**
	 * Creates and sets up the layout for this window.
	 */
	private void initUI() {
		final GUI copygui = this;
		this.setTitle("Master");
		this.setSize(800, 600);
		this.setExtendedState(Frame.MAXIMIZED_BOTH);
		this.setLocationRelativeTo(null);
		this.setDefaultCloseOperation(EXIT_ON_CLOSE);
		
		JPanel panel = new JPanel();
		JPanel tablePanel = new JPanel();
		
		panel.setLayout(new FlowLayout());
		panel.setPreferredSize(new Dimension(Frame.MAXIMIZED_HORIZ, 75));
		
		this.setLayout(new BorderLayout());
		
		this.getContentPane().add(panel,BorderLayout.NORTH);
		this.getContentPane().add(tablePanel,BorderLayout.CENTER);
		
		JButton quitButton = new JButton("Quit");
		quitButton.setToolTipText("Quits the program");
		quitButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				jobs.shutDown();
				workers.shutDown();
				
				System.exit(0);
			}
		});

		JButton loadApplicationsButton = new JButton("Load Applications");
		loadApplicationsButton.setToolTipText("Loads applications from a file");
		loadApplicationsButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent event) {
				try {
					loadApplicationsFromFile("applications.txt");
					loadInsFromFile("ins.txt");
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		});
		
		JButton loadWorkersButton = new JButton("Load Workers");
		loadWorkersButton.setToolTipText("Loads some workers from a file");
		loadWorkersButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				try {
					workers.clear();
					loadWorkersFromFile("workers.txt");
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		});

		JButton loadJobsButton = new JButton("Load Jobs");
		loadJobsButton.setToolTipText("Loads jobs from a file");
		loadJobsButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent event) {
				try {
					loadJobsFromFile("jobs.txt");
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		});
		
		JButton startButton = new JButton("Start");
		startButton.setToolTipText("Starts the worker computation");
		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				workers.start();
			}
		});
		
		JButton changeSchedulingButton = new JButton("Change Scheduling Algorithm");
		changeSchedulingButton.setToolTipText("Change Scheduling Algorithm");
		changeSchedulingButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				new AlgorithmFrame(copygui);
			}
		});
		
		JButton updateWorkerButton = new JButton("Update Worker Details");
		updateWorkerButton.setToolTipText("Updates worker details");
		updateWorkerButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				workers.update();
			}
		});
		
		JButton addJobButton = new JButton("Add Job");
		addJobButton.setToolTipText("Adds a new job");
		addJobButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				new AddJobFrame(copygui);
			}
		});
		
		JButton cancelJobButton = new JButton("Cancel Job");
		cancelJobButton.setToolTipText("Cancels the selected job");
		cancelJobButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				if (jtJobs.getSelectedRowCount() == 0) {
					showWarning("You haven't selected a job.");
				} else {
					int id = jobs.cancelJob(jtJobs.getSelectedRow());
					
					if (id != -1) {
						workers.cancelJob(id);
					}
				}
			}
		});

		JButton deleteJobButton = new JButton("Delete Job");
		deleteJobButton.setToolTipText("Deletes the selected job");
		deleteJobButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				if (jtJobs.getSelectedRowCount() == 0)
					showWarning("You haven't selected a job.");
				else
					jobs.deleteJob(jtJobs.getSelectedRow());
			}
		});

		JButton rescheduleJobButton = new JButton("Re-schedule Job");
		rescheduleJobButton.setToolTipText("Re-schedules the selected job");
		rescheduleJobButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				if (jtJobs.getSelectedRowCount() == 0)
					showWarning("You haven't selected a job.");
				else
					if (!jobs.rescheduleJob(jtJobs.getSelectedRow()))
						showWarning("That job is already scheduled.");
			}
		});
		
		Dimension preferred = new Dimension(600, 400);
		
		jtWorkers = new JTable(workers);
		jtWorkers.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		JScrollPane jspWorkers = new JScrollPane(jtWorkers);
		jspWorkers.setPreferredSize(preferred);

		jtJobs = new JTable(jobs);
		jtJobs.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		JScrollPane jspJobs = new JScrollPane(jtJobs);
		jspJobs.setPreferredSize(preferred);
		
		jtApplications = new JTable(applications);
		JScrollPane jspApps = new JScrollPane(jtApplications);
		jspApps.setPreferredSize(new Dimension(430, 200));

		panel.add(quitButton);
		panel.add(loadApplicationsButton);
		panel.add(loadWorkersButton);
		panel.add(loadJobsButton);
		panel.add(startButton);
		panel.add(changeSchedulingButton);
		panel.add(updateWorkerButton);
		panel.add(addJobButton);
		panel.add(cancelJobButton);
		panel.add(deleteJobButton);
		panel.add(rescheduleJobButton);
		tablePanel.add(jspWorkers);
		tablePanel.add(jspJobs);
		tablePanel.add(jspApps);
	}
	
	/**
	 * Displays a message box dialog to the user containing the supplied String.
	 * @param warning The String to display in the message box.
	 */
	private void showWarning(String warning) {
		JOptionPane.showMessageDialog((JFrame)this, warning, "Warning",
				JOptionPane.WARNING_MESSAGE);
	}
	
	/**
	 * Loads a list of available applications that can be scheduled to run on
	 * the Workers. 
	 * @param filename The file containing the list of applications.
	 * @throws IOException if there is any problem reading from the file.
	 */
	public void loadApplicationsFromFile(String filename) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(filename));
		String line = null;
		ArrayList<String> apps = new ArrayList<String>();
		String[] result = new String[0];
		
		while ((line = br.readLine()) != null) {
			apps.add(line);
			applications.newApp(line);
		}
		
		result = apps.toArray(result);
		workers.setExecs(result);
	}

	/**
	 * Loads a list of available in files that can be used by
	 * the Workers. 
	 * @param filename The file containing the list of in files.
	 * @throws IOException if there is any problem reading from the file.
	 */
	public void loadInsFromFile(String filename) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(filename));
		String line = null;
		ArrayList<String> ins = new ArrayList<String>();
		String[] result = new String[0];
		
		while ((line = br.readLine()) != null) {
			ins.add(line);
			applications.newIn(line);
		}
		
		result = ins.toArray(result);
		workers.setIns(result);
	}
	
	/**
	 * From a file, creates a number of new WorkerHandler objects used to
	 * interact with the Worker clients. On creation, each WorkerHandler tries
	 * to send a Setup Worker command to its Worker.
	 * @param filename The file containing the list of Workers.
	 * @throws IOException if there is any problem reading from the file.
	 */
	public void loadWorkersFromFile(String filename) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(filename));
		String line = null;
		String lines[] = null;
		String address = null;
		int port = -1; 
		String password = null;
		int cost = -1;
		
		while ((line = br.readLine()) != null) {
			lines = line.split(",");
			address = lines[0];
			port = Integer.parseInt(lines[1]);
			password = lines[2];
			cost = Integer.parseInt(lines[3]);
			
			workers.newWorker(address, port, password, cost);
		}
		
		for(int i = 0; i < 7; i++) {
			packColumn(jtWorkers, i, 1);
		}
	}
	
	/**
	 * From a file, creates a number of new Job objects.
	 * @param filename The file containing the list of Jobs.
	 * @throws IOException if there is any problem reading from the file.
	 */
	public void loadJobsFromFile(String filename) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(filename));
		String line = null;
		String application = null;
		String applicationArgs = null;
		Date deadline = null;
		String lines[] = null;
		int budget;
		int priority;
		SimpleDateFormat format = new SimpleDateFormat("yyyy.MM.dd.HH.mm.ss");
		
		while ((line = br.readLine()) != null) {
			lines = line.split(",");
			application = lines[0];
			applicationArgs = lines[1];
			try {
				deadline = format.parse(lines[2]);
			} catch (ParseException e) {
				e.printStackTrace();
			}
			budget = Integer.parseInt(lines[3]);
			priority = Integer.parseInt(lines[4]);
			
			jobs.newJob(application, applicationArgs, deadline, budget, priority);
		}
		
		for(int i = 0; i < 6; i++) {
			packColumn(jtJobs, i, 1);
		}
	}
	
	/**
	 * Public method to allow the adding of new Jobs.
	 */
	public void addJob(String application, String applicationargs, Date deadline,
			int budget, int priority){
		jobs.newJob(application, applicationargs, deadline, budget, priority);
	}

	/**
	 * Public method to allow the adding of new Jobs.
	 */
	public void addJob(String application, String applicationargs, Date deadline,
			int budget, int priority, Date delay){
		jobs.newJob(application, applicationargs, deadline, budget, priority,
						delay);
	}
	
	/**
	 * Change the scheduling algorithm used by the Master when scheduling Jobs.
	 * @param algo The scheduling algorithm to change to.
	 */
	public void changeAlgorithm(JobManager.SchedulingAlgorithm algo){
		jobs.changeSchedulingAlgorithm(algo);
	}
	
	/**
	 * Auto-sizes columns of a JTable to be the size of its biggest cell.
	 */
	public static void packColumn(JTable table, int vColIndex, int margin) {
		table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

	    DefaultTableColumnModel colModel =
	    	(DefaultTableColumnModel)table.getColumnModel();
	    TableColumn col = colModel.getColumn(vColIndex);
	    int width = 0;

	    // Get width of column header
	    TableCellRenderer renderer = col.getHeaderRenderer();
	    if (renderer == null) {
	        renderer = table.getTableHeader().getDefaultRenderer();
	    }
	    java.awt.Component comp = renderer.getTableCellRendererComponent(
	        table, col.getHeaderValue(), false, false, 0, 0);
	    width = comp.getPreferredSize().width;

	    // Get maximum width of column data
	    for (int r=0; r<table.getRowCount(); r++) {
	        renderer = table.getCellRenderer(r, vColIndex);
	        comp = renderer.getTableCellRendererComponent(
	            table, table.getValueAt(r, vColIndex), false, false, r, vColIndex);
	        width = Math.max(width, comp.getPreferredSize().width);
	    }

	    // Add margin
	    width += 2 * margin;

	    // Set the width
	    col.setPreferredWidth(width);
	}	
}