import java.util.Date;

public class Job implements Comparable<Job>  {
	private int id;
	private String application;
	private String applicationArgs;
	private Date deadline;
	private int priority;
	private int budget;
	private JobStatus status;

	private boolean hasDelay;
	private Date delay;

	/**
	 * The differenct statuses a Job can have.
	 */
	public enum JobStatus {
		QUEUED,
		COMPUTING,
		COMPLETE,
		FAILED
	}
	
	/**
	 * Job Constructor.
	 * Made default constructor private so that it can't be called
	 */
	@SuppressWarnings("unused")
	private Job() {
	}
	
	/**
	 * Job Constructor.
	 * @param id Job ID. Should be unique among all the Jobs in the JobManager
	 * 			 collection.
	 * @param application The application to be run for this Job.
	 * @param applicationArgs The application arguments to be supplied to the
	 * 						  application.
	 * @param deadline The deadline for when the Job needs to be finished by.
	 * @param budget The budget required by this Job.
	 * @param priority The priority of this Job. A lower value means a higher
	 * 				   priority.
	 */
	public Job(int id, String application, String applicationArgs, Date deadline,
			int budget, int priority) {
		this.id = id;
		this.application = application;
		this.applicationArgs = applicationArgs;
		this.deadline = deadline;
		this.budget = budget;
		this.priority = priority;
		
		this.status = JobStatus.QUEUED;
	}
	
	/**
	 * Job Constructor.
	 * @param id Job ID. Should be unique among all the Jobs in the JobManager
	 * 			 collection.
	 * @param application The application to be run for this Job.
	 * @param applicationArgs The application arguments to be supplied to the
	 * 						  application.
	 * @param deadline The deadline for when the Job needs to be finished by.
	 * @param budget The budget required by this Job.
	 * @param priority The priority of this Job. A lower value means a higher
	 * 				   priority.
	 * @param delay The delay time before this Job should start.
	 */
	public Job(int id, String application, String applicationArgs, Date deadline,
			int budget, int priority, Date delay) {
		this.id = id;
		this.application = application;
		this.applicationArgs = applicationArgs;
		this.deadline = deadline;
		this.budget = budget;
		this.priority = priority;

		this.hasDelay = true;
		this.delay = delay;
		
		this.status = JobStatus.QUEUED;
	}
	
	/**
	 * Any shutdown code that may be required when the program is exiting should
	 * go here.
	 */
	public void shutdown() {
	}
	
	/**
	 * Getter for id of Job.
	 */
	public int getId() {
		return this.id;
	}
	
	/**
	 * Getter for application of Job.
	 */
	public String getApp() {
		return this.application;
	}
	
	/**
	 * Getter for arguments supplied to the application of Job.
	 */
	public String getArgs() {
		return this.applicationArgs;
	}
	
	/**
	 * Getter for deadline of Job.
	 */
	public Date getDeadline() {
		return this.deadline;
	}
	
	/**
	 * Getter for budget of Job.
	 */
	public int getBudget() {
		return this.budget;
	}
	
	/**
	 * Getter for priority of Job.
	 * If the delay hasn't passed yet, set priority to be very low.
	 */
	public int getPriority() {
		if ((this.hasDelay) && (this.delay.after(new Date()))) {
			return 99;
		}
		return this.priority;
	}
	
	/**
	 * Getter for status of Job.
	 */
	public JobStatus getStatus() {
		return this.status;
	}
	
	/**
	 * Setter for status of Job.
	 */
	public void setStatus(JobStatus status) {
		this.status = status;
	}

	/**
	 * Need to override this method to allow us to sort Jobs by priority, then
	 * deadline and then finally id.
	 * @param other The other job to compare to.
	 * @return A negative integer, zero, or a positive integer as this Job is
	 * less than, equal to, or greater than the specified Job. 
	 */
	@Override
	public int compareTo(Job other) {
		int output;
		if ((output = this.getPriority() - other.getPriority()) == 0) {
			if ((output = this.deadline.compareTo(other.getDeadline())) == 0) {
				output = this.id - other.getId();
			}
		}
		
		return output;
	}
	
	/**
	 * toString method used to display Job information on the command-line.
	 */
	@Override
	public String toString() {
		return "Job " + this.id + ":\t\t PRIORITY:" + this.getPriority() +
			"\t\tDEADLINE: " + this.deadline;
	}
}