import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import javax.swing.table.AbstractTableModel;

public class JobManager extends AbstractTableModel implements Runnable {
	private static final long serialVersionUID = 2923684122525319153L;
	
	private ArrayList<Job> jobsQueue;
	private ArrayList<Job> jobsComputing;
	private ArrayList<Job> jobsComplete;
	private ArrayList<Job> jobsFailed;
	
	private int numJobs;
	private int currentId;
	
	private int updateFrequency;
	
	private SchedulingAlgorithm sa;
	
	private Comparator<WorkerHandler> comp;
	
	private Thread t;
	
	/**
	 * The different scheduling algorithms that can be used.
	 */
	public enum SchedulingAlgorithm {
		FIFO,
		EDF,
		EDCBB
	}

	/**
	 * JobManager Constructor.
	 * Passes 10 seconds as the updateFrequency to the 1-parameter Constructor.
	 */
	public JobManager() {
		this(10000);
	}
	
	/**
	 * JobManager Constructor.
	 * Initialises the required variables and starts the Thread related to this
	 * class.
	 * @param updateFrequency How often the JobManager table should be refreshed.
	 */
	public JobManager(int updateFrequency) {
		this.jobsQueue = new ArrayList<Job>();
		this.jobsComputing = new ArrayList<Job>();
		this.jobsComplete = new ArrayList<Job>();
		this.jobsFailed = new ArrayList<Job>();
		
		this.numJobs = 0;
		this.currentId = 0;
		
		this.updateFrequency = updateFrequency;
		
		this.sa = SchedulingAlgorithm.FIFO;
		
		this.comp = null;
		
		this.t = new Thread(this);
		this.t.start();
	}

	/**
	 * Call the shutdown method of each Job stored within the JobManager.
	 */
	public void shutDown() {
		ArrayList<Job> js = getJobs();
		
		for (Job j : js)
			j.shutdown();
	}

	/**
	 * Changes the scheduling algorithm by selecting which Comparator to use
	 * when sorting.
	 * @param sa The SchedulingAlgorithm to use. 
	 */
	public void changeSchedulingAlgorithm(SchedulingAlgorithm sa) {
		this.sa = sa;
		
		switch (sa) {
		case FIFO:
			comp = null;
			break;
		case EDF:
			comp = new WorkerTimeComparator();
			break;			
		case EDCBB:
			comp = new WorkerCostComparator();
			break;
		}
		
		sortJobsQueue();
	}

	/**
	 * Getter for the scheduling algorithm.
	 */
	public SchedulingAlgorithm getSchedulingAlgorithm() {
		return this.sa;
	}
	
	/**
	 * If not using the first-come-first-serve scheduling algorithm then sort
	 * the Jobs Queue.
	 */
	public void sortJobsQueue() {
		if (sa != SchedulingAlgorithm.FIFO) {
			Collections.sort(jobsQueue);
			this.fireTableRowsUpdated(0, this.numJobs - 1);
		}
	}
	
	/**
	 * Getter for the WorkerHandler Comparator.
	 */
	public Comparator<WorkerHandler> getWorkerComparator () {
		return comp;
	}
	
	/**
	 * Setter for the update frequency.
	 */
	public void setUpdateFrequency(int updateFrequency) {
		this.updateFrequency = updateFrequency;
	}
	
	/**
	 * Create a new Job (with out a delay) and add it to the Jobs Queue.
	 */
	public void newJob(String application, String applicationArgs, Date deadline,
			int budget, int priority) {
		Job j = new Job(currentId, application, applicationArgs, deadline,
				budget, priority);
		
		jobsQueue.add(j);
		numJobs++;
		currentId++;
		
		fireTableRowsInserted(numJobs, numJobs);
	}

	/**
	 * Create a new Job (with a delay) and add it to the Jobs Queue.
	 */
	public void newJob(String application, String applicationArgs, Date deadline,
			int budget, int priority, Date delay) {
		Job j = new Job(currentId, application, applicationArgs, deadline,
				budget, priority, delay);
		
		jobsQueue.add(j);
		numJobs++;
		currentId++;
		
		fireTableRowsInserted(numJobs, numJobs);
	}
	
	/**
	 * Cancel a Job.
	 * @param row The row in the table where the Job to be cancelled is shown.
	 */
	public int cancelJob(int row) {
		ArrayList<Job> js = getJobs();
		Job j = js.get(row);
		int id = j.getId();
		int idReturn = -1;
		boolean found = false;

		for (Job j2 : jobsQueue) {
			if (j2.getId() == id) {
				jobsQueue.remove(j2);
				j2.setStatus(Job.JobStatus.FAILED);
				jobsFailed.add(j2);
				found = true;
				break;
			}
		}
		
		if (!found) {
			for (Job j2 : jobsComputing) {
				if (j2.getId() == id) {
					jobsComputing.remove(j2);
					j2.setStatus(Job.JobStatus.FAILED);
					jobsFailed.add(j2);
					found = true;
					idReturn = id;
					break;
				}
			}
		}
		
		this.fireTableRowsUpdated(0, this.numJobs - 1);
		
		return idReturn;
	}
	
	/**
	 * Delete a Job.
	 * @param row The row in the table where the Job to be deleted is shown.
	 */
	public void deleteJob(int row) {
		ArrayList<Job> js = getJobs();
		Job j = js.get(row);
		int id = j.getId();
		boolean found = false;

		for (Job j2 : jobsQueue) {
			if (j2.getId() == id) {
				jobsQueue.remove(j2);
				found = true;
				break;
			}
		}
		
		if (!found) {
			for (Job j2 : jobsComputing) {
				if (j2.getId() == id) {
					jobsComputing.remove(j2);
					found = true;
					break;
				}
			}
		}
		
		if (!found) {
			for (Job j2 : jobsComplete) {
				if (j2.getId() == id) {
					jobsComplete.remove(j2);
					found = true;
					break;
				}
			}
		}

		if (!found) {
			for (Job j2 : jobsFailed) {
				if (j2.getId() == id) {
					jobsFailed.remove(j2);
					found = true;
					break;
				}
			}
		}
		
		if (found) {
			numJobs--;
			this.fireTableRowsDeleted(row, row);
		}
	}
	
	/**
	 * 
	 * @param row The row in the table where the Job to be rescheduled is shown.
	 * @return True if the Job is already Queued or currently computing.
	 */
	public boolean rescheduleJob(int row) {
		ArrayList<Job> js = getJobs();
		Job j = js.get(row);
		int id = j.getId();
		boolean found = false;
		
		for (Job j2 : jobsComplete) {
			if (j2.getId() == id) {
				jobsComplete.remove(j2);
				j2.setStatus(Job.JobStatus.QUEUED);
				jobsQueue.add(j2);
				found = true;
				break;
			}
		}
	

		if (!found) {
			for (Job j2 : jobsFailed) {
				if (j2.getId() == id) {
					jobsFailed.remove(j2);
					j2.setStatus(Job.JobStatus.QUEUED);
					jobsQueue.add(j2);
					found = true;
					break;
				}
			}
		}
		
		if (found) {
			this.fireTableRowsUpdated(0, this.numJobs - 1);
		} 
		
		return found;
	}
	
	/**
	 * Returns all the Jobs stored in JobManager.
	 */
	public ArrayList<Job> getJobs() {
		ArrayList<Job> js = new ArrayList<Job>();
		
		for (Job j : jobsQueue)
			js.add(j);
		
		for (Job j : jobsComputing)
			js.add(j);
			
		for (Job j : jobsComplete)
			js.add(j);

		for (Job j : jobsFailed)
			js.add(j);
		
		return js;
	}
	
	/**
	 * Re-adds a Job to the Jobs Queue.
	 */
	public void readdJob(Job j) {
		jobsQueue.add(j);
		return;
	}
	
	/**
	 * Adds a Job to the Computing Jobs list.
	 */
	public void computeJob(Job j) {
		jobsComputing.add(j);
		return;
	}
	
	/**
	 * Returns the next Job in the Jobs Queue.
	 */
	public Job getNextJob() {
		Job result = null;

		if (!jobsQueue.isEmpty()) {
			result = jobsQueue.remove(0);
		}
		
		return result;
	}
	
	/**
	 * Changes a Job's status.
	 * @param j The Job to modify.
	 * @param status The new status to give to the Job.
	 */
	public void changeJobStatus(Job j, Job.JobStatus status) {
		j.setStatus(status);
		
		if (status == Job.JobStatus.COMPLETE) {
			jobsComputing.remove(j);
			jobsComplete.add(j);
		}
		
		this.fireTableRowsUpdated(0, this.numJobs - 1);
	}
	
	/**
	 * Returns the Job Queue count.
	 */
	public int getJobQueueCount() {
		return jobsQueue.size();
	}

	/**
	 * Returns true is all Jobs are complete or have failed.
	 */
	public boolean getAllJobsComplete() {
		return ((jobsQueue.size() + jobsComputing.size()) == 0);
	}
	
	/**
	 * Returns a name for the column.
	 * @param col The column being queried.
     * @return A string containing the name of column.
	 */
	@Override
	public String getColumnName(int col) {
		switch (col) {
		case 0:
			return "Job ID";
		case 1:
			return "Application";
		case 2:
			return "Applcation Arguments";
		case 3:
			return "Deadline";
		case 4:
			return "Budget";
		case 5:
			return "Priority";
		case 6:
			return "Status";
		}
		throw new IllegalArgumentException("Bad column (" + col + ")");
	}
	
	/**
	 * Returns the number of columns in the model. A JTable uses this method to
	 * determine how many columns it should create and display.
     * @return The number of columns in the model.
	 */
	@Override
	public int getColumnCount() {
		return 7;
	}

	/**
	 * Returns the number of rows in the model.
	 * A JTable uses this method to determine how many rows it should display.
     * @return The number of rows in the model.
	 */
	@Override
	public int getRowCount() {
		return numJobs;
	}

	/**
	 * Returns the value for the cell for a given row and column.
	 * @param row The row whose value is to be queried.
	 * @param col The column whose value is to be queried.
     * @return The value Object at the specified cell.
	 */
	@Override
	public Object getValueAt(int row, int col) {
		switch (col){
		case 0:
			return getJobs().get(row).getId();
		case 1:
			return getJobs().get(row).getApp();
		case 2:
			return getJobs().get(row).getArgs();
		case 3:
			return getJobs().get(row).getDeadline();
		case 4:
			return getJobs().get(row).getBudget();
		case 5:
			return getJobs().get(row).getPriority();
		case 6:
			return getJobs().get(row).getStatus();
		}
		throw new IllegalArgumentException("Bad cell (" + row + ", " + col +
											")");
	}

	/**
	 * The Thread loop that updates the table every <updateFrequency> ms.
	 */
	@Override
	public void run() {
		while(true) {
			this.fireTableRowsUpdated(0, this.numJobs - 1);
			
			try {
				Thread.sleep(updateFrequency);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Returns a multi-line string containing all the results of calling 
	 * toString on all the Jobs contained in the JobManager.
	 */
	@Override
	public String toString() {
		String result = "";
		
		for (Job j : getJobs())
			result = result + j.toString() + "\n";
		
		return result;
	}
}