import java.io.IOException;
import java.util.Date;

public class WorkerHandler {
	private int id;
	private int port;
	private int jobID;
	private Boolean alive;
	private Boolean hasApps;
	private String addr;
	private String pass;
	private Connection connection;
	private WorkerStatus currentStatus;
	private Job currentJob;
	
	private int cost;
	private long timeTaken;
	private Date startTime;
	private int numJobs;
	
	/**
	 * The different commands that can be sent to a Worker.
	 */
	public enum Command {
		SETUP (-1),
		STATUS (0),
		NEWJOB (1),
		CANCEL (2),
		END (9);
		
		private final int code;
		
		Command(int code) {
			this.code = code;
		}
		
		public int getCode() {
			return code;
		}
	}
	
	/**
	 * The different statuses that a Worker can have.
	 */
	public enum WorkerStatus {
		WAITING (0),
		BUSY (1),
		FINISHED (2),
		DEAD (3);
		
		private final int code;
		
		WorkerStatus(int code) {
			this.code = code;
		}
		
		public int getCode() {
			return code;
		}
	}
	
	/**
	 * WorkerHandler Constructor.
	 * Made default constructor private so that it can't be called
	 */
	@SuppressWarnings("unused")
	private WorkerHandler() {
	}
	
	/**
	 * WorkerHandler Constructor.
	 * @param id WorkerHandler ID. Should be unique among all the WorkerHandler
	 * 			 in the WorkerManager collection.
	 * @param addr IP address of the Worker.
	 * @param port Port used to connect to the Worker.
	 * @param password Password used to authenticate with the Worker.
	 * @param cost The cost of the Worker.
	 */
	public WorkerHandler(int id, String addr, int port, String password,
							int cost) {
		this.id = id;
		this.addr = addr;
		this.port = port;
		this.pass = password;
		this.alive = true;
		this.hasApps = false;
		this.jobID = -1;
		
		this.connection = null;
		
		this.currentStatus = null;
		
		this.currentJob = null;

		this.cost = cost;
		this.timeTaken = -1;
		this.startTime = null;
		this.numJobs = 0;
	}
	
	/**
	 * Any shutdown code that may be required when the program is exiting should
	 * go here.
	 */
	public void shutdown() {
		comCancel();
	}
	
	/**
	 * Attempts to connect to the Worker and get its status.
	 * @return The status of the Worker.
	 * @throws IOException if there is a problem with the connection.
	 */
	public int getStatus() throws IOException {
		this.connection.sendInt(Command.STATUS.getCode());
		int status = this.connection.recvInt();
		if (status < 0) {
			throw new IOException("Error recieved from Worker");
		} else if (status == 2) {
			this.currentJob = null;
			this.connection.setBusy();
			this.connection.recvFile(this.jobID + ".out");
			while (this.connection.isBusy()) { }
		}
		return status;
	}
	
	/**
	 * Attempts to connect to the Worker and send it the application files and
	 * input files.
	 * @param execFiles The applications to send.
	 * @param inputFiles The input files to send.
	 */
	public void comSetup(String[] execFiles, String[] inputFiles) {
		int i, status;
		int numExec = execFiles.length;
		int numInput = inputFiles.length;
		
		if (alive) {
			try {
				this.connection = new Connection(this.addr,this.port);
				this.connection.sendText(this.pass);
				status = getStatus();
				
				if (status == 0) {
					this.connection.sendInt(Command.SETUP.getCode());
					
					this.connection.sendInt(numExec);
					this.connection.sendInt(numInput);
					
					for (i=0; i< numExec; i++) {
						this.connection.sendText(execFiles[i]);
						this.connection.sendFile(execFiles[i]);
					}
					
					for (i=0; i < numInput; i++) {
						this.connection.sendText(inputFiles[i]);
						this.connection.sendFile(inputFiles[i]);
					}
					
					this.hasApps = true;
				}
				
				this.connection.sendInt(Command.END.getCode());
				this.connection.closeSocket();
			} catch (Exception e) {
				System.err.println(addr + " comSetup:");
				System.err.println("\t" + e.getMessage());
				this.setDead();
			} finally {
				this.connection = null;
			}
		} 
		
		return;
	}
	
	/**
	 * Attempts to connect to the Worker and get its status.
	 * @return The status of the Worker.
	 */
	public int comStatus() {
		int status = -1;
		
		try {
			this.connection = new Connection(this.addr,this.port);
			this.connection.sendText(this.pass);
			status = getStatus();
			
			this.connection.sendInt(Command.END.getCode());
			this.connection.closeSocket();
			
			this.alive = true;
		} catch (Exception e) {
			System.err.println(addr + " comStatus:");
			System.err.println("\t" + e.getMessage());
			this.setDead();
			status = -1;
		} finally {
			this.connection = null;
		}
	
		return status;
	}
	
	/**
	 * Attempts to connect to the Worker and give it a new Job.
	 * @param job The Job to send to the Worker.
	 * @return Whether the Worker successfully started the Job.
	 */
	public int comNew(Job job) {
		int success = 0, status;
		
		if (alive) {
			try {
				this.connection = new Connection(this.addr,this.port);
				this.connection.sendText(this.pass);
				status = getStatus();
				
				if (status == 0) {
					this.currentJob = job;
					
					this.connection.sendInt(Command.NEWJOB.getCode());
					
					this.connection.sendText(job.getApp());
					this.connection.sendText(job.getArgs());
					
					success = this.connection.recvInt();
				}
				
				this.connection.sendInt(Command.END.getCode());
				this.connection.closeSocket();
			} catch (Exception e) {
				System.err.println(e.getMessage());
				this.setDead();
				success = 0;
			} finally {
				this.connection = null;
			}
		}
		
		if (success == 1) {
			this.jobID = job.getId();
		}
		
		return success;
	}
	
	/**
	 * Attempts to connect to the Worker and cancel a currently running Job.
	 * @return Whether the cancel was a success.
	 */
	public int comCancel() {
		int success = -1, status;
		
		if (alive) {
			this.jobID = -1;
			
			try {
				this.connection = new Connection(this.addr,this.port);
				this.connection.sendText(this.pass);
				status = getStatus();
				
				if (status == 1) {
					this.connection.sendInt(2);
					success = this.connection.recvInt();
				}
				
				this.connection.sendInt(Command.END.getCode());
				this.connection.closeSocket();
			} catch (Exception e) {
				System.err.println(e.getMessage());
				this.setDead();
				success = 0;
			} finally {
				this.connection = null;
			}
		}
		
		return success;
	}
	
	/**
	 * Gets the status of the Worker by connecting to it.
	 */
	public WorkerStatus printStatus() {
		Date now = new Date();
				
		switch(this.comStatus()) {
		case 0:
			this.currentStatus = WorkerStatus.WAITING;
			
			if (this.numJobs > 0) {
				this.startTime = now;
			}
			break;
		case 1:
			this.currentStatus = WorkerStatus.BUSY;
			break;
		case 2:
			this.currentStatus = WorkerStatus.FINISHED;
			
			if (this.timeTaken == -1 && this.startTime != null) {
				this.timeTaken = now.getTime() - this.startTime.getTime();
			}
			
			break;
		default:
			this.currentStatus = WorkerStatus.DEAD;
			break;
		}
		
		return this.currentStatus;
	}

	/**
	 * Gets the status of the Worker by not connecting to it and just using a 
	 * previously saved status.
	 */
	public WorkerStatus getCurrentStatus() {
		return this.currentStatus;
	}

	/**
	 * Getter for the id of WorkerHandler.
	 */
	public int getId() {
		return this.id;
	}

	/**
	 * Getter for the current Job id of WorkerHandler.
	 */
	public int getJobId() {
		return this.jobID;
	}

	/**
	 * Getter for the IP address of WorkerHandler.
	 */
	public String getAddress() {
		return this.addr;
	}

	/**
	 * Getter for whether the WorkerHandler has been sent the applications and
	 * input files.
	 */
	public boolean getHasApps() {
		return this.hasApps;
	}

	/**
	 * Getter for the current Job of WorkerHandler.
	 */
	public Job getCurrentJob() {
		return this.currentJob;
	}

	/**
	 * Getter for the time taken of WorkerHandler.
	 */
	public long getTimeTaken() {
		return this.timeTaken;
	}

	/**
	 * Getter for the cost of WorkerHandler.
	 */
	public int getCost() {
		return this.cost;
	}

	/**
	 * Getter for the Job number of WorkerHandler.
	 */
	public int getJobNum() {
		return this.numJobs;
	}

	/**
	 * Setter for the Job of WorkerHandler.
	 */
	public void setJob(Job next) {
		this.currentJob = next;
	}

	/**
	 * Setter for whether this WorkerHandler is currently dead or not.
	 */
	public void setDead() {
		this.alive = false;
	}
	
	/**
	 * Schedule the number of Jobs that this Worker can take on.
	 * If supplied with no arguments then schedule no Jobs.
	 * @return The number of Jobs this Worker can take on.
	 */
	public int schedule() {
		this.numJobs = 0;
		
		return this.numJobs;
	}

	/**
	 * Schedule the number of Jobs that this Worker can take on.
	 * @param finishTime The estimated finish time of the current batch of Jobs.
	 * @return The number of Jobs this Worker can take on.
	 */
	public int schedule(Date finishTime) {
		long time;
		this.numJobs = 1;
		
		if (this.timeTaken > -1) {
			time = finishTime.getTime() - this.startTime.getTime();
			numJobs = (new Double(Math.floor(time / this.timeTaken))).intValue();
		}
		
		return this.numJobs;
	}
	
	/**
	 * Schedule the number of Jobs that this Worker can take on.
	 * @param maxTime Of all the Workers, the max time taken.
	 * @return The number of Jobs this Worker can take on.
	 */
	public int schedule(long maxTime) {
		this.numJobs = 1;
		
		if (this.timeTaken > -1) {
			numJobs = Math.round((float)maxTime / (float)this.timeTaken);
		}
		
		return this.numJobs;
	}

	/**
	 * Returns a String that lets the user know that a Job has been assigned
	 * to a Worker.
	 */
	public String printConfirm() {
		return "> Job #" + this.currentJob.getId() + " assigned to Worker #" +
			this.id;
	}

	/**
	 * Returns a String used for displaying a Worker's details to the 
	 * command-line.
	 */
	@Override
	public String toString() {
		String time = "";
		String cost = null;
		
		String start = "Worker #" + this.id + ":\tSTATUS: " + this.printStatus();
		
		if (this.currentStatus == WorkerStatus.FINISHED) {
			cost  = "\tCOST: " + this.cost;
		} else {
			cost  = "\t\tCOST: " + this.cost;
		}
		
		if (this.timeTaken == -1) {
			time  = "\t\tTIME TAKEN: NONE";
		} else {
			time  = "\t\tTIME TAKEN: " +
				(new Double(Math.floor(this.timeTaken)).intValue());
		}
		
		String jobs  = "\tNUM JOBS QUEUED: " + this.numJobs;
		return start + cost + time + jobs;
	}
}