import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import javax.swing.table.AbstractTableModel;

public class WorkerManager extends AbstractTableModel implements Runnable {
	private static final long serialVersionUID = 8547063980498329886L;
	
	private ArrayList<WorkerHandler> workers;
	private JobManager jobs;
	
	private boolean doWork;
	
	private int numWorkers;
	
	private int workingUpdateFrequency;
	private int idleUpdateFrequency;
	
	private String[] execs = {};
	private String[] ins = {};
	
	private long maxTime;
	
	private Thread t;
	
	/**
	 * WorkerManager Constructor.
	 * Made default constructor private so that it can't be called
	 */
	@SuppressWarnings("unused")
	private WorkerManager() {
	}
	
	/**
	 * WorkerManager Constructor.
	 * Passes 10 seconds as the idleUpdateFrequency and 1 second as the
	 * workingUpdateFrequency to the 3-parameter Constructor.
	 * @param jobs
	 */
	public WorkerManager(JobManager jobs) {
		this(jobs, 10000, 1000);
	}
	
	/**
	 * WorkerManager Constructor.
	 * Initialises the required variables and starts the Thread related to this
	 * class.
	 * @param jobs The JobManager created by the Master class.
	 * @param idleUpdateFrequency How often the WorkerManager table should be
	 * 							  refreshed while not computing Jobs.
	 * @param workingUpdateFrequency How often the WorkerManager table should be
	 * 								 refreshed while computing Jobs.
	 */
	public WorkerManager(JobManager jobs, int idleUpdateFrequency,
			int workingUpdateFrequency) {
		this.workers = new ArrayList<WorkerHandler>();
		
		this.jobs = jobs;
		
		this.doWork = false;
		
		this.numWorkers = 0;
		
		this.idleUpdateFrequency = idleUpdateFrequency;
		this.workingUpdateFrequency = workingUpdateFrequency;
		
		this.maxTime = 0;
		
		this.t = new Thread(this);
		this.t.start();
	}
	
	/**
	 * Clears the ArrayList and Table of Workers.
	 */
	public void clear() {
		this.numWorkers = 0;
		workers.clear();
	}
	
	/**
	 * Starts the assigning of Jobs to Workers through a check in the run()
	 * method.
	 */
	public void start() {
		this.doWork = true;
	}
	
	/**
	 * Calls the shutdown() method for each of the Workers contained in the 
	 * WorkerManager.
	 */
	public void shutDown() {
		for (WorkerHandler worker : getWorkers())
			worker.shutdown();
	}
	
	/**
	 * Refreshed the WorkerManager table.
	 */
	public void update() {
		this.fireTableRowsUpdated(0, numWorkers - 1);
	}
	
	/**
	 * Setter for the applications that are allowed to be used by Jobs.
	 */
	public void setExecs(String[] execs) {
		this.execs = execs;
	}

	/**
	 * Returns a multi-line String containing all the applications that are
	 * allowed to be used by Jobs.
	 */
	public String printExecs() {
		String result = "";
		
		for (int i = 0; i < this.execs.length; i++)
			result = result + execs[i] + "\n";
		
		return result;
	}
	
	/**
	 * Setter for the input files that are allowed to be used by Jobs.
	 */
	public void setIns(String[] ins) {
		this.ins = ins;
	}
	
	/**
	 * Creates a new WorkerHandler and updates the WorkManager Table.
	 */
	public void newWorker(String address, int port, String password, int cost) {
		WorkerHandler w = new WorkerHandler(numWorkers, address, port, password,
				cost);
		w.comSetup(execs, ins);
		w.printStatus();
		
		workers.add(w);
		numWorkers ++;
		
		fireTableRowsInserted(numWorkers, numWorkers);
	}
	
	/**
	 * Returns all the WorkerHandlers that are stored within the WorkerManager.
	 */
	public ArrayList<WorkerHandler> getWorkers() {
		return workers;
	}
	
	/**
	 * Gets a WorkerHandler to send the Cancel Job command to its Worker.
	 * @param id The id of the WorkerHandler to get to send the Cancel command.
	 */
	public void cancelJob(int id) {
		for (WorkerHandler w : workers) {
			if (w.getJobId() == id) {
				w.comCancel();
			}
		}
	}
	
	/**
	 * Returns a name for the column.
	 * @param col The column being queried.
     * @return A string containing the name of column.
	 */
	@Override
	public String getColumnName(int col) {
		switch (col){
		case 0:
			return "Worker ID";
		case 1:
			return "IP Address";
		case 2:
			return "Status";
		case 3:
			return "Current Job ID";
		case 4:
			return "Cost";
		case 5:
			return "Time Taken";
		case 6:
			return "Num Jobs Queued";
		}
		throw new IllegalArgumentException("Bad column (" + col + ")");
	}
	
	/**
	 * Returns the number of columns in the model. A JTable uses this method to
	 * determine how many columns it should create and display.
     * @return The number of columns in the model.
	 */
	@Override
	public int getColumnCount() {
		return 7;
	}

	/**
	 * Returns the number of rows in the model.
	 * A JTable uses this method to determine how many rows it should display.
     * @return The number of rows in the model.
	 */
	@Override
	public int getRowCount() {
		return numWorkers;
	}

	/**
	 * Returns the value for the cell for a given row and column.
	 * @param row The row whose value is to be queried.
	 * @param col The column whose value is to be queried.
     * @return The value Object at the specified cell.
	 */
	@Override
	public Object getValueAt(int row, int col) {
		switch (col){
		case 0:
			return getWorkers().get(row).getId();
		case 1:
			return getWorkers().get(row).getAddress();
		case 2:
			return getWorkers().get(row).getCurrentStatus();
		case 3:
			if (getWorkers().get(row).getJobId() == -1)
				return "-";
			else
				return getWorkers().get(row).getJobId();
		case 4:
			return getWorkers().get(row).getCost();
		case 5:
			return getWorkers().get(row).getTimeTaken();
		case 6:
			return getWorkers().get(row).getJobNum();
		}
		throw new IllegalArgumentException("Bad cell (" + row + ", " + col +
											")");
	}

	/**
	 * The Thread loop that updates the table every <idleUpdateFrequency> ms
	 * or <workingUpdateFrequency> ms. If the Master is in the working state 
	 * every <workingUpdateFrequency> ms the WorkerManager will check the status
	 * of every WorkerHandler and if one is free, assign a Job to it.
	 */
	@Override
	public void run() {
		while(true) {
			if (this.doWork) {
				if (jobs.getAllJobsComplete()) {
					this.doWork = false;
				} else {
					checkForAvailableWorkers();
				}
				
				try {
					Thread.sleep(workingUpdateFrequency);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			} else {
				for (WorkerHandler w : workers)
					w.printStatus();
				this.fireTableRowsUpdated(0, numWorkers - 1);
				
				try {
					Thread.sleep(idleUpdateFrequency);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * Checks to see if there are any Workers free to do work.
	 */
	private void checkForAvailableWorkers() {
		ArrayList<WorkerHandler> free = new ArrayList<WorkerHandler>();
		Iterator<WorkerHandler> wit = null;
		Iterator<WorkerHandler> fit = null;
		WorkerHandler w = null;
		WorkerHandler.WorkerStatus status = null;
		Comparator<WorkerHandler> comp = jobs.getWorkerComparator();
		int assigned = 0;
		
		for (int i = 0; i < workers.size(); i++) {
			w = workers.get(i);
			status = w.printStatus();
			
			if (status == WorkerHandler.WorkerStatus.WAITING) {
				free.add(w);
				
				if (w.getTimeTaken() > maxTime) {
					maxTime = w.getTimeTaken();
				}
			} else if (status == WorkerHandler.WorkerStatus.FINISHED) {
				jobs.changeJobStatus(w.getCurrentJob(), Job.JobStatus.COMPLETE);
				this.fireTableRowsUpdated(i, i);
			} else {
				this.fireTableRowsUpdated(i, i);
			}
		}
		
		assigned = jobs.getJobQueueCount();
		if (free.size() > 0) {
			if (comp != null) {
				Collections.sort(workers, comp);
				Collections.sort(free, comp);
			}
			jobs.sortJobsQueue();
			
			wit = workers.iterator();
			while (wit.hasNext()) {
				w = wit.next();
				if (assigned > 0) {
					assigned -= w.schedule(maxTime);
				} else {
					w.schedule();
				}
			}
			
			fit = free.iterator();
			while (fit.hasNext()) {
				w = fit.next();
				if (w.getJobNum() > 0 || jobs.getSchedulingAlgorithm() ==
										JobManager.SchedulingAlgorithm.FIFO) {
					workerAvailable(w);
				}
			}
		}
	}

	/**
	 * Called when there is a Worker available. Assigns the next Job in the 
	 * queue which is determined by the scheduling algorithm.
	 * @param worker WorkerHandler of the available Worker.
	 */
	private void workerAvailable(WorkerHandler worker) {
		ArrayList<Job> unused = new ArrayList<Job>();
		boolean capable = false;
		Job j = null;
		int success = -1;
		
		if (jobs.getSchedulingAlgorithm() == JobManager.SchedulingAlgorithm.EDCBB
				&& worker.getTimeTaken() != -1) {
			while (!capable) {
				j = jobs.getNextJob();
				if (j != null) {
					if (j.getBudget() < (worker.getCost()*worker.getTimeTaken())) {
						capable = true;
					} else {
						unused.add(j);
						j = null;
					}
				} else {
					break;
				}
			}
		} else {
			j = jobs.getNextJob();
		}
		
		Iterator<Job> uit = unused.iterator();
		while (uit.hasNext()) {
			jobs.readdJob(uit.next());
		}
		
		if (j != null) {			
			jobs.computeJob(j);
			
			success = worker.comNew(j);
			if (success == 1) {
				jobs.changeJobStatus(j, Job.JobStatus.COMPUTING);
			} else {
				jobs.changeJobStatus(j, Job.JobStatus.FAILED);
			}
			this.fireTableRowsUpdated(0, this.numWorkers - 1);
			jobs.fireTableDataChanged();
		} else {
			this.doWork = false;
		}
	}
	
	/**
	 * Returns a multi-line String of the result of every WorkerHandler's
	 * toString() method.
	 */
	@Override
	public String toString() {
		String result = "";
		
		for (WorkerHandler worker : workers)
			result = result + worker.toString() + "\n";
		
		return result;
	}
}