import java.io.*;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.net.InetAddress;

public class Worker {
	int port, status, jobID, error;
	ArrayList<String> apps;
	String key, name, outFile;
	ServerConnection connection;
	boolean connected, blocked;
	Job job;
	
	/** Constructor for Worker Class
	 * 
	 * @param name		Worker name
	 * @param port		Port for server to listen on
	 * @param password	Password for verification
	 */
	public Worker(String name, int port, String password) { 
		this.name = name;
		this.port = port;
		this.key = password;
		this.connected = false;
		this.blocked = false;
		
		this.error = 0;
		this.status = 0;
		this.job = null;
		this.jobID = 0;
		this.apps = new ArrayList<String>();
		
		try {
			this.connection = new ServerConnection(this.port);
		} catch (Exception e) {
			e.printStackTrace();
			this.blocked = true;
		}
	}
	
	/** Getter for blocked boolean
	 * 
	 * @return NOT connected blocked
	 */
	public boolean notBlocked() {
		return !this.blocked;
	}
	
	/** Getter for connected boolean
	 * 
	 * @return NOT connected boolean
	 */
	public boolean notConnected() {
		return !this.connected;
	}
	
	/**	Verifies the password with the given key made on initialisation
	 * 
	 * 	@return If it passes the password check
	 */
	private boolean checkPass(String password) {
		return password.equalsIgnoreCase(key);
	}
	
	private void openConnection() throws IOException {
	    this.connected = true;
		this.connection.openSocket();
	}
	
	/**	Closes the current socket connection
	 * 
	 */
	private void closeConnection() {
	    if (this.connection != null) {
	        this.connection.closeSocket();
	    }
	    this.connected = false;
	}
	
	/**	Run method for Worker
	 * Loops a listening socket while Worker is not broken
	 * 
	 */
	public void run() {
		SimpleDateFormat sdf = new SimpleDateFormat("[HH:mm:ss] ");
		String timestamp = "";
		boolean loop = true;
		int command;
		String password;
		try {
		    openConnection();
		    password = this.connection.recvText();
			if (checkPass(password)) {
				while (loop) {
					command = this.connection.recvInt();
					timestamp = sdf.format(new Date());
					switch (command) {
					case -1:			// Setup Worker
						System.out.println(timestamp + "SETUP recieved from Master");
						comSetup();
						break;
					case 0:				// Check Status
						System.out.println(timestamp + "STATUS recieved from Master");
						comStatus();
						break;
					case 1:				// New Job
						System.out.println(timestamp + "NEW JOB recieved from Master");
						comNew();
						break;
					case 2:				// Cancel Job
						System.out.println(timestamp + "CANCEL recieved from Master");
						comCancel();
						break;
					case 9:
						loop = false;
						break;
					default:
						throw new IOException(timestamp + "Unknown command '" + command + "' recieved." );
					}
				}
				this.error -= 1;
			}
		} catch (Exception e) {
			e.printStackTrace();
			this.error++;
			
			if (error > 5) {
			    this.blocked = true;
			}
		} finally {
			closeConnection();
		}
	}
	
	/**	Response to Setup Command.
	 *  Worker waits for number of files, then filename file pairs
	 * 
	 * @throws IOException Thrown on socket error
	 */
	private void comSetup() throws IOException {
		int i, numExec, numInput;
		String fileName;
		
		numExec = this.connection.recvInt();
		numInput = this.connection.recvInt();
		
		for (i=0; i < numExec; i++) {
			fileName = this.connection.recvText();
			this.apps.add(fileName);
			System.out.println("APP #" + i + " '" + fileName + "' from Master");
			this.connection.recvFile(fileName);
		}
		
		for (i=0; i < numInput; i++) {
			fileName = this.connection.recvText();
			System.out.println("FILE #" + i + " '" + fileName + "' from Master");
			this.connection.setBusy();
			this.connection.recvFile(fileName);
			while (this.connection.isBusy()) {}	
		}
	}
	
	/**	Response to Status Check.
	 * 	Simple response with status. Also checks if output is ready to be sent
	 * 
	 * @throws IOException
	 */
	private void comStatus() throws IOException {
		if (this.status > 0) {
			this.status = job.getStatus(); // get status from job (-1 - Error, 0 - Waiting, 1 - Running, 2 - Finished)
		}
		
		this.connection.sendInt(this.status);
		
		// If job has finished, send output
		if (this.status == 2) {	
			this.sendOut();
		}
	}
	
	/**	Sends output file over the connection
	 * 
	 * @throws IOException	Thrown on socket errors
	 */
	private void sendOut() throws IOException {
		System.out.println("SENDING '" + this.outFile + "' output file");
		this.jobID++;
		this.connection.setBusy();
		this.connection.sendFile(this.outFile); // get filename from job;
		while (this.connection.isBusy()) {
		}
		
		this.status = 0;
	}
	
	/**	Response to NewJob command.
	 *  Creates and starts a new job thread which runs external executables
	 * 
	 * @throws IOException	Thrown by socket errors
	 */
	private void comNew() throws IOException {
		int success = 0;
		String appName, appArgs;
				
		appName = this.connection.recvText();
		appArgs = this.connection.recvText();
		
		this.outFile = this.name + "_" + jobID + ".out";
		
		if (this.status == 0) {
			if (this.apps.contains(appName)) {
				job = new Job(jobID, appName, appArgs, this.outFile);
				job.start();
				this.status = 1;
				success = 1;
			} else {
				System.out.println("APP '" + appName + "' not found.");
			}
		}
		this.connection.sendInt(success);
	}
	
	/** Response to a cancel command.
	 *  Cancels the current job running.
	 * 
	 * @throws IOException Thrown on socket error
	 */
	private void comCancel() throws IOException {
		this.status = job.getStatus();
		
		if (this.status == 1) {
			job.stopJob();
			this.status = 0;
		}
		
		this.connection.sendInt(status);
	}
	
	/** Main method that drives the Worker
	 * 
	 * @param args	Program arguments
	 */
	public static void main(String[] args) {
		try {
		    String name;
    	    int port;
    	    String password;
    	    String addr;
    	    
			if (args.length < 3) {
			    name = "default";
			    port = 1234;
			    password = "password";
			} else {
			    name = args[0];
			    port = Integer.parseInt(args[1]);
			    password = args[2];
			}
			
			addr = InetAddress.getLocalHost().toString();
			
			System.out.println("Starting Server");
			System.out.println("Addr: \t" + addr);
			System.out.println("Port: \t" + port + "\n");
			System.out.println("Listening...");
			
			Worker worker = new Worker(name, port, password);
			       
			while (worker.notBlocked()) {
				if (worker.notConnected()) {
					worker.run();
				}
			}
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
